from ordeq import node
from ordeq.framework import get_node
from ordeq_common import StringBuffer

x = StringBuffer("x")
y = StringBuffer("y")
z = StringBuffer("z")
one = StringBuffer("1")


@node(inputs=[x, y], outputs=[z, one], tags=[])
def func1(x: str, y: str) -> tuple[str, str]:
    return f"{x} + {y}", y


node1 = get_node(func1)
print(node1.tags)


@node(inputs=[x, y], outputs=[z, one], tags=["tag1", "tag2"])
def func2(x: str, y: str) -> tuple[str, str]:
    return f"{x} + {y}", y


node2 = get_node(func2)
print(node2)
print(node2.tags)


@node(inputs=[x, y], outputs=[z, one], tags={"key1": "value1"})
def func3(x: str, y: str) -> tuple[str, str]:
    return f"{x} + {y}", y


node3 = get_node(func3)
print(node3.tags)


@node(inputs=[x, y], outputs=[z, one], tags=None)
def func4(x: str, y: str) -> tuple[str, str]:
    return f"{x} + {y}", y


node4 = get_node(func4)
print(node4.tags)



from dataclasses import dataclass
from pathlib import Path

from ordeq import IO


@dataclass(kw_only=True, frozen=True)
class ExampleIO(IO[str]):
    path: Path
    attribute: str

    def load(self, hello: str = "Hello") -> str:
        return f"{self.path}@{self.attribute}: {hello} world!"

    def save(self, data: str, hello: str = "Hello") -> None:
        assert data == f"{self.path}@{self.attribute}: {hello} world!"


# Load:
example_input = ExampleIO(path=Path("hello.txt"), attribute="L1")
# No kwarg:
print(example_input.load())
# Alternative kwarg:
print(example_input.load(hello="Hello there"))

with_load_options = example_input.with_load_options(hello="Hey")
# No kwarg, with load options:
print(with_load_options.load())
# Alternative kwarg, with load options:
print(with_load_options.load(hello="Hi"))
print(type(with_load_options))

# Save:
example_output = ExampleIO(path=Path("world.txt"), attribute="L2")
# No kwarg:
example_output.save("world.txt@L2: Hello world!", hello="Hello")
# Alternative kwarg:
example_output.save("world.txt@L2: Guten tag world!", hello="Guten tag")

with_save_options = example_output.with_save_options(hello="Hey")
# No kwarg, with save options:
with_save_options.save("world.txt@L2: Hey world!")
# Alternative kwarg, with save options:
with_save_options.save("world.txt@L2: Hi world!", hello="Hi")
print(type(with_save_options))



from ordeq import node
from ordeq.framework.runner import run
from ordeq_common import Static, StringBuffer

x1 = Static(1)
x2 = StringBuffer()
x3 = StringBuffer("2")
x4 = StringBuffer()


@node(inputs=x1, outputs=x2)
def increment(x: int) -> str:
    return f"{x + 1}"


@node(inputs=[x2, x3], outputs=x4)
def decrement(x: str, y: str) -> str:
    return f"{int(x) - int(y)}"


regular = run(increment, decrement, verbose=True)

print(regular)

# provide alternative IO when running the pipeline
patched = run(
    increment,
    decrement,
    io={x1: Static(2), x3: Static("33"), x4: StringBuffer()},
    verbose=True,
)

print(patched)



from dataclasses import dataclass
from pathlib import Path

from ordeq import Input


@dataclass(kw_only=True, frozen=True)
class ExampleInputLoadArg(Input[str]):
    path: Path
    attribute: str

    def load(self) -> int:
        return 3


example_input = ExampleInputLoadArg(path=Path("hello.txt"), attribute="L1")
print(example_input.load())



from ordeq.framework.hook import Hook, InputHook, NodeHook, OutputHook, RunHook
from ordeq.framework.io import IO, Input, IOException, Output
from ordeq.framework.nodes import Node, NodeNotFound, get_node, node
from ordeq.framework.pipeline import Pipeline, is_pipeline
from ordeq.framework.runner import run

__all__ = [
    "IO",
    "Hook",
    "IOException",
    "Input",
    "InputHook",
    "Node",
    "NodeHook",
    "NodeNotFound",
    "Output",
    "OutputHook",
    "Pipeline",
    "RunHook",
    "get_node",
    "is_pipeline",
    "node",
    "run",
]



from ordeq import Output, OutputHook


class MyInvalidTypedOutputHook(OutputHook[str]):
    def before_output_save(
        self, op: Output[bytes], data: bytes
    ) -> None:
        print(f"saving data `{data}` to output `{op}`")


_ = MyInvalidTypedOutputHook()



from types import ModuleType

from ordeq import node, run

from resources.catalog.catalogs import local, remote

catalog: ModuleType = local


@node(inputs=catalog.hello, outputs=catalog.result)
def func1(hello: str) -> str:
    return f"{hello.upper()}!"


print(run(func1))

catalog: ModuleType = remote


@node(inputs=catalog.hello, outputs=catalog.result)
def func2(hello: str) -> str:
    return f"{hello.upper()}!"


print(run(func2))



from time import time, sleep
from ordeq import RunHook
from ordeq import node, run
from ordeq_common import StringBuffer


class RunTimer(RunHook):
    start_time: float

    def before_run(self, graph):
        self.start_time = time()

    def after_run(self, graph, data):
        end_time = time()
        elapsed_time = end_time - self.start_time
        print(f"Total run time: {elapsed_time:.1f} seconds")


x = StringBuffer()
y = StringBuffer()


@node(outputs=x)
def func1() -> str:
    return "Hello"


@node(outputs=y)
def func2() -> str:
    return "world"


run(func1, func2, hooks=[RunTimer()])



from ordeq import node

from resources.catalog.catalogs import inconsistent

catalog = inconsistent


@node(inputs=catalog.hello, outputs=catalog.result)
def func(hello: str) -> str:
    return f"{hello.upper()}!"



from ordeq import node
from ordeq_common import Static, StringBuffer


@node(inputs=(StringBuffer("a"), Static(value=4)), outputs=StringBuffer("z"))
def func(*args: str | int) -> str:
    return "".join(str(i) for i in args)


print(func("a", 4))



from dataclasses import dataclass

from ordeq import Input


@dataclass(kw_only=True, frozen=True)
class ExampleInputLoadArg(Input):
    def load(self) -> str:
        return "Hello world"


_ = ExampleInputLoadArg().with_load_options(hello="hello world")



from collections.abc import Callable
from typing import TypeGuard

Pipeline = set[Callable]

"""
A pipeline represents a set of nodes. Example usage:

```python
>>> # in reference/to.py
>>> from ... import node1, node2
>>> pipeline: Pipeline = {node1, node2}
```

Pipelines can be used to run a set of nodes through the Ordeq CLI, like so:

```shell
$ python {your-entrypoint} run \
    --pipeline {reference}.{to}:{pipeline}
```

You can also run multiple pipelines together, like so:

```shell
$ python {your-entrypoint} run \
    --pipeline {reference}.{to}:{pipeline1} \
    {reference}.{to}:{pipeline2}
```

Ordeq will resolve the dependencies between nodes in the pipelines.

Lastly, it is possible to run a combination of nodes and pipelines:

```shell
$ python {your-entrypoint} run \
    --pipeline {reference}.{to}:{pipeline} \
    --node {reference}.{to}:{node}
```

You can also run the pipeline programmatically as follows:

```python
>>> # in reference/to.py
>>> from reference.to import pipeline
>>> from ordeq.framework.runner import run
>>> run(*pipeline)
```

Pipelines are sets, so any set operation works on pipelines:

```python
>>> run(*pipeline1.union(pipeline2))
```

"""


def is_pipeline(obj: object) -> TypeGuard[Pipeline]:
    """Returns whether an object is a pipeline, meaning a set of nodes. Used
    because ``isinstance`` cannot be used on generic type ``set[Node]``.

    Args:
        obj: the object

    Returns:
        whether ``obj`` is a pipeline or not
    """

    return isinstance(obj, set) and all(callable(o) for o in obj)



from ordeq import node
from ordeq_common import StringBuffer


@node(
    inputs=[StringBuffer("x"), StringBuffer("y")],
    outputs=[StringBuffer("z"), StringBuffer("1")],
)
def func(x: str, y: str) -> tuple[str, str]:
    """A really nice node"""

    return f"{x} + {y}", y


print(func.__doc__)
print(func.__annotations__)
print(func.__name__)
print(func.__module__)



from ordeq import node
from ordeq_common import StringBuffer

mock_x = StringBuffer("X")
mock_z = StringBuffer("Z")


def func(x: str = "X") -> str:
    return x


a = node(func=func, inputs=[mock_x], outputs=[mock_z])
b = node(func=a, inputs=[mock_x], outputs=[mock_z])
assert a("X") == b("X")
assert a != b



from dataclasses import dataclass
from pathlib import Path

from ordeq import IO


class ExampleStaticIO(IO[str]):
    @staticmethod
    def load() -> str:
        return "loaded"

    @staticmethod
    def save(data: str) -> None:
        print(data)


# This currently raises an exception because the first argument of `save` needs
# to be `self`.
example_io = ExampleStaticIO()
print(example_io.load())
print(example_io.save("saved"))



from types import ModuleType
from unittest.mock import Mock

import pytest
from ordeq.framework.catalog import CatalogError, check_catalogs_are_consistent
from ordeq_common import Static, StringBuffer


@pytest.mark.parametrize(
    ("a", "b"),
    [
        (
            {
                # empty
            },
            {
                # empty
            },
        ),
        (
            {
                # empty
            },
            {
                "something_else": 4  # not an IO
            },
        ),
        ({"hello": StringBuffer()}, {"hello": Static("hello")}),
        (
            {"hello": StringBuffer(), "result": StringBuffer()},
            {
                "hello": Static("hello"),
                "result": StringBuffer(),
                "something_else": 4,  # not an IO
            },
        ),
    ],
)
def test_it_checks_consistent(a, b):
    catalog_a = Mock(ModuleType)
    catalog_b = Mock(ModuleType)
    catalog_a.__dict__ = a
    catalog_b.__dict__ = b
    check_catalogs_are_consistent(catalog_a, catalog_b)


@pytest.mark.parametrize(
    ("a", "b"),
    [
        (
            {"hello": StringBuffer(), "result": StringBuffer()},
            {
                "hello": Static("hello")
                # result is mising
            },
        ),
        (
            {
                "hello": StringBuffer()
                # result is missing
            },
            {"hello": Static("hello"), "result": StringBuffer()},
        ),
        (
            {
                # empty
            },
            {"hello": Static("hello")},
        ),
        (
            {
                # empty but contains non-IO
                "something_else": 4
            },
            {"hello": Static("hello"), "result": StringBuffer()},
        ),
    ],
)
def test_it_checks_inconsistent(a, b):
    catalog_a = Mock(ModuleType)
    catalog_b = Mock(ModuleType)
    catalog_a.__dict__ = a
    catalog_b.__dict__ = b
    with pytest.raises(CatalogError, match="Catalogs are inconsistent"):
        check_catalogs_are_consistent(catalog_a, catalog_b)



# Imported by ./run_modules.py
from ordeq import node
from ordeq_common import Static, StringBuffer

x1 = Static(12345)
x2 = StringBuffer()
x3 = StringBuffer("12345")
x4 = StringBuffer()


@node(inputs=x1, outputs=x2)
def increment(x: int) -> str:
    return f"{x + 1}"


@node(inputs=[x2, x3], outputs=x4)
def decrement(x: str, y: str) -> str:
    return f"{int(x) - int(y)}"



import logging

from ordeq import node
from ordeq.framework import get_node
from ordeq.framework.graph import NodeGraph
from ordeq.framework.runner import _run_graph
from ordeq_common import StringBuffer

logging.basicConfig(level=logging.INFO)
A, B, C, D, E, F = [StringBuffer(c) for c in "ABCDEF"]

plus = node(func=lambda x, y: f"{x} + {y}", inputs=(A, B), outputs=(C,))
minus = node(func=lambda x, y: f"{x} - {y}", inputs=(C, D), outputs=(E,))
square = node(func=lambda x: f"({x})^2", inputs=(E,), outputs=(F,))

nodes = [get_node(n) for n in (plus, minus, square)]
_run_graph(NodeGraph.from_nodes(nodes))



from collections.abc import Generator
from typing import IO, Protocol, runtime_checkable

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self


@runtime_checkable
class PathLike(Protocol):
    """Utility protocol used for the type of the `path` attribute of
    file-like datasets. Example types that adhere to this protocol are
    `pathlib.Path` and `cloudpathlib.CloudPath`:

    ```python
    >>> from ordeq.types import PathLike
    >>> from cloudpathlib import CloudPath
    >>> isinstance(CloudPath, PathLike)
    True
    >>> from pathlib import Path
    >>> isinstance(Path, PathLike)
    True

    ```

    `pathlib.Path` provides a consistent file addressing mechanism across
    different operating systems. Moreover, `cloudpathlib.CloudPath` allows
    users to interact with files on cloud storage through (most of the)
    `pathlib.Path` API.  The `PathLike` protocol allows users to use
    either one interchangebly in the dataset definition.

    Note that datasets that rely on the IO implementation of an external
    library may expect a type different from `PathLike` for
    its `path` (e.g., `spark.read_csv`).
    """

    def open(
        self, mode="r", buffering=-1, encoding=None, errors=None, newline=None
    ) -> IO: ...

    def __fspath__(self): ...


class GlobPath(PathLike, Protocol):
    """Protocol for accepting any PathLike object that has a `glob` method"""

    def glob(
        self, pattern, *, case_sensitive=None
    ) -> Generator[Self, None, None]: ...



from dataclasses import dataclass
from pathlib import Path

from ordeq import Input


@dataclass(kw_only=True, frozen=True)
class ExampleInputLoadKwarg(Input):
    path: Path
    attribute: str

    def load(self, hello: str = "Hi") -> str:
        return f"{self.path}@{self.attribute}: {hello} world!"


example_input = ExampleInputLoadKwarg(path=Path("hello.txt"), attribute="L1")
# No kwarg:
print(example_input.load())
# Alternative kwarg:
print(example_input.load(hello="Hello"))
print(type(example_input))

with_options = example_input.with_load_options(hello="Hello")
# No kwarg, with load options:
print(with_options.load())
# Alternative kwarg, with load options:
print(with_options.load(hello="Hello"))
print(type(with_options))


@dataclass(kw_only=True, frozen=True)
class ExampleInputLoadArg(Input):
    path: Path
    attribute: str

    def load(self, hello: str = "Hello") -> str:
        return f"{self.path}@{self.attribute}: {hello} world!"


example_input_arg = ExampleInputLoadArg(path=Path("hello.txt"), attribute="L1")
print(example_input_arg.load(hello="Hello"))
print(type(example_input_arg))

with_options_arg = example_input_arg.with_load_options(hello="Hello")
# This should raise a type error,
# but still run as we fill the missing kwarg on load:
# TODO: Add a check during with_load_options to ensure all required load_args are set.
print(with_options_arg.load())
# Alternative kwarg, with load options:
print(with_options_arg.load(hello="Hi"))
print(type(with_options_arg))
# Unknown kwarg:
example_input_arg.with_load_options(unknown_kwarg="Hello")  # should error



import pytest
from ordeq_common import StringBuffer


@pytest.fixture(autouse=True)
def doctest_setup(doctest_namespace):
    """Mock the example datasets that are used in the doctests."""
    doctest_namespace["CSV"] = StringBuffer()
    doctest_namespace["Table"] = StringBuffer()



import pytest
from ordeq import Node, node
from ordeq.framework import IO, get_node
from ordeq.framework.nodes import NodeNotFound
from ordeq.framework.runner import _run_node
from ordeq_common.io.string_buffer import StringBuffer


class TestNode:
    @pytest.mark.parametrize(
        ("kwargs", "expected"),
        [({"x": "_X_"}, "_X_ + Y"), ({"x": "_X_", "y": "_Y_"}, "_X_ + _Y_")],
    )
    def test_it_runs_with_defaulted_args(self, kwargs, expected):
        @node(inputs=StringBuffer("x"), outputs=StringBuffer("z"))
        def func(x: str, y: str = "Y") -> str:
            return f"{x} + {y}"

        actual = func(**kwargs)

        assert actual == expected

    @pytest.mark.parametrize(
        ("inputs", "expected"),
        [
            ((), ""),
            ((StringBuffer("a"),), "a"),
            ((StringBuffer("a"), StringBuffer("b")), "ab"),
        ],
    )
    def test_it_runs_with_args(self, inputs, expected):
        @node(inputs=inputs, outputs=StringBuffer("z"))
        def func(*args: str) -> str:
            return "".join(args)

        actual = func(*(i.load() for i in inputs))

        assert actual == expected


@pytest.mark.parametrize(
    "inputs",
    [
        (),  # too few
        (StringBuffer("a"), StringBuffer("b")),  # 1 too many
        (
            StringBuffer("a"),
            StringBuffer("b"),
            StringBuffer("c"),
        ),  # 2 too many
    ],
)
def test_it_raises_for_invalid_inputs(inputs: tuple[IO]):
    with pytest.raises(  # noqa: PT012
        ValueError, match="Node inputs invalid for function arguments"
    ):
        n = Node(func=lambda _: _, inputs=inputs, outputs=(StringBuffer("c"),))
        _run_node(n, hooks=())


def method_w_0_ret(a: str) -> None:
    return


def method_w_1_ret(a: str) -> str:
    return a


def method_w_2_ret(a: str) -> tuple[str, str]:
    return a, a


@pytest.mark.parametrize(
    ("func", "outputs"),
    [
        (method_w_0_ret, (StringBuffer("a"), StringBuffer("b"))),  # 1 too many
        (
            method_w_0_ret,
            (StringBuffer("a"), StringBuffer("b"), StringBuffer("c")),
        ),  # 2 too many
        (method_w_1_ret, ()),  # too few
        (method_w_1_ret, (StringBuffer("a"), StringBuffer("b"))),  # 1 too many
        (
            method_w_1_ret,
            (StringBuffer("a"), StringBuffer("b"), StringBuffer("c")),
        ),  # 2 too many
        (method_w_2_ret, ()),  # too few
        (method_w_2_ret, (StringBuffer("a"),)),  # too few
        (
            method_w_2_ret,
            (StringBuffer("a"), StringBuffer("b"), StringBuffer("c")),
        ),  # 1 too many
    ],
)
def test_it_raises_for_invalid_outputs(func, outputs: tuple[IO]):
    with pytest.raises(  # noqa: PT012
        ValueError, match="Node outputs invalid for return annotation"
    ):
        n = Node(func=func, inputs=(StringBuffer("a"),), outputs=outputs)
        _run_node(n, hooks=())


class TestGetNode:
    def test_it_gets_registered_node(self):
        @node(inputs=[StringBuffer("a")], outputs=[StringBuffer("b")])
        def my_func(a: str) -> str:
            return a

        node_obj = get_node(my_func)
        assert node_obj is not None
        assert node_obj.func == my_func

    def test_it_raises_node_not_found(self):
        def my_func(a: str) -> str:
            return a

        with pytest.raises(NodeNotFound, match="Node 'my_func' not found"):
            get_node(my_func)



from dataclasses import dataclass

from ordeq import Input, node
from ordeq_common import StringBuffer


@dataclass(frozen=True)
class Unhashable(Input[list]):
    # This input is unhashable because its data attribute is a list.
    data: list

    def load(self) -> list:
        return self.data


@node(inputs=[Unhashable(["y", "z"])], outputs=StringBuffer("y"))
def func(x: str) -> str:
    return x



from ordeq import node
from ordeq_common import StringBuffer

mock_x = StringBuffer("X")
mock_y = StringBuffer("Y")
mock_z = StringBuffer("Z")


@node(inputs=[mock_x, mock_y], outputs=[mock_z])
def func(x: str = "X", y: str = "Y") -> str:
    return x + y


print(func("x", "y"))
print(func("x"))
print(func(x="x", y="y"))
print(func("x", y="y"))



from random import shuffle

from ordeq import IO, node, run

o1 = IO()
o2 = IO()
o3 = IO()
o4 = IO()


@node(outputs=o1)
def f1(): ...


@node(outputs=o2)
def f2(): ...


@node(outputs=o3)
def f3(): ...


@node(outputs=o4)
def f4(): ...


@node(inputs=[o1, o2, o3, o4])
def a(x1, x2, x3, x4): ...


@node(inputs=[o1, o2, o3, o4])
def z(x1, x2, x3, x4): ...


pipeline = {f1, f2, f3, f4, a, z}
# shuffle
x = list(pipeline)
shuffle(x)
pipeline = set(x)

run(*pipeline, verbose=True)



from dataclasses import dataclass
from pathlib import Path

from ordeq import Input


@dataclass(kw_only=True, frozen=True)
class ExampleInputLoadArg(Input):
    path: Path
    attribute: str

    def load(self, hello: str = "...") -> str:
        """My docstring"""
        return f"{self.path}@{self.attribute}: {hello} world!"


example_input = ExampleInputLoadArg(path=Path("hello.txt"), attribute="L1")
print(example_input.load("Hello"))



from ordeq import node
from ordeq_common import StringBuffer, Static

x = Static("X")


@node(inputs=x, outputs=x)  # outputs should be of type Output or IO
def func(data: str) -> str:
    return data + data



from unittest.mock import Mock

import pytest
from ordeq.framework.graph import (
    _build_graph,
    _find_sink_nodes,
    _find_topological_ordering,
    _nodes,
)
from ordeq_common import StringBuffer

A, B, C, D, E, F = [StringBuffer(c) for c in "ABCDEF"]


@pytest.mark.parametrize(
    ("edges", "expected"),
    [
        ({"A": ["B"], "B": []}, {"B"}),
        ({"A": ["B", "C"], "B": [], "C": [], "D": ["A"]}, {"B", "C"}),
        ({"A": ["B"], "B": ["A"]}, set()),
        ({"A": [], "B": [], "C": []}, {"A", "B", "C"}),
        ({}, set()),
    ],
)
def test_find_sink_nodes(edges, expected):
    assert _find_sink_nodes(edges) == expected


def test_it_builds_a_graph():
    first = Mock()
    first.name = "first"
    first.inputs = [A]
    first.outputs = [B]

    second = Mock()
    second.name = "second"
    second.inputs = [B, C]
    second.outputs = [D]

    third = Mock()
    third.name = "third"
    third.inputs = [B, D]
    third.outputs = [F]

    edges = _build_graph([third, second, first])
    assert edges == {first: [third, second], second: [third], third: []}
    assert _nodes(edges) == {first, second, third}


def test_it_builds_graph_with_single_node():
    first = Mock()
    first.name = "first"
    first.inputs = [A]
    first.outputs = [B]

    edges = _build_graph([first])
    assert edges == {first: []}
    assert _nodes(edges) == {first}


def test_it_raises_error_on_duplicated_outputs():
    first = Mock()
    first.name = "first"
    first.inputs = [A]
    first.outputs = [B]

    second = Mock()
    second.name = "second"
    second.inputs = [D]
    second.outputs = [B]

    with pytest.raises(
        ValueError, match="cannot be outputted by more than one node"
    ):
        _build_graph([first, second])


@pytest.mark.parametrize(
    ("edges", "expected"),
    [
        (
            # Example 0:
            # 0 ---
            # |     |
            # 1 --- 2
            {0: [1, 2], 1: [2], 2: []},
            (0, 1, 2),
        ),
        (
            # Example 1:
            # 0 --- 2
            # |     |
            # 1 --- 3
            {0: [1, 2], 1: [3], 2: [3], 3: []},
            (0, 2, 1, 3),
        ),
        (
            # Example 2:
            # 0 --- 2
            # |
            # 1
            {0: [1, 2], 1: [], 2: []},
            (0, 2, 1),
        ),
        (
            # Example 3:
            # 0 --- 2
            # |  /  |
            # 1 --- 3
            {0: [1, 2], 1: [2, 3], 2: [3], 3: []},
            (0, 1, 2, 3),
        ),
        (
            # Example 4:
            # 0 --- 2
            # |  X  |
            # 1 --- 3
            {0: [1, 2, 3], 1: [2, 3], 2: [3], 3: []},
            (0, 1, 2, 3),
        ),
    ],
)
def test_it_finds_a_topological_ordering(edges, expected):
    # Note that even though for some of the examples there exist
    # multiple valid topological orderings, the ordering is
    # deterministic (as dictionaries are ordered).

    actual = _find_topological_ordering(edges)
    assert actual == expected



from dataclasses import dataclass
from pathlib import Path

from ordeq import Output


@dataclass(kw_only=True, frozen=True)
class ExampleOutputsaveArg(Output[int]):
    path: Path
    attribute: str

    def save(self, data: str) -> None:
        assert data == f"{self.path}@{self.attribute}: Hello world!"


example_output = ExampleOutputsaveArg(path=Path("hello.txt"), attribute="L1")
example_output.save("hello.txt@L1: Hello world!")



from dataclasses import dataclass

from ordeq import IO, Input, Output


@dataclass(kw_only=True, frozen=True)
class NameGenerator(Input[str]):
    name: str

    def load(self) -> str:
        return self.name


@dataclass(kw_only=True, frozen=True)
class NamePrinter(Output[str]):
    def save(self, data: str) -> None:
        print(f"Name: {data}")


@dataclass(kw_only=True, frozen=True)
class SayHello(IO[str]):
    name: Input[str]
    writer: Output[str]

    def load(self) -> str:
        return f"Hello, {self.name.load()}!"

    def save(self, data: str) -> None:
        self.writer.save(data)


name_generator = NameGenerator(name="John")
print(name_generator)
print(name_generator.references)

name_printer = NamePrinter()
print(name_printer)
print(name_printer.references)

message = SayHello(name=name_generator, writer=name_printer)
print(message)
print(message.references)

assert name_generator != message



import os

from ordeq import node, run

from resources.catalog.catalogs import local, remote

os.environ["ENV"] = "local"


def get_catalog():
    return local if os.environ["ENV"] == "local" else remote


catalog = get_catalog()


@node(inputs=catalog.hello, outputs=catalog.result)
def func1(hello: str) -> str:
    return f"{hello.upper()}!"


print(run(func1))

os.environ["ENV"] = "acceptance"
catalog = get_catalog()


@node(inputs=catalog.hello, outputs=catalog.result)
def func2(hello: str) -> str:
    return f"{hello.upper()}!"


print(run(func2))



from ordeq import Input, InputHook, Node, Output, OutputHook, node, run
from ordeq_common import StringBuffer


class MyInputHook(InputHook[str]):
    def before_input_load(self, io: Input[str]) -> None:
        print("Before loading data from:", io)

    def after_input_load(
        self, io: Input[str], data: str
    ) -> None:
        print("After loading data from:", io)


class MyOutputHook(OutputHook[str]):
    def before_output_save(
        self, io: Output[str], data: str
    ) -> None:
        print("Before saving data to:", io)

    def after_output_save(
        self, io: Output[str], data: str
    ) -> None:
        print("After saving data to:", io)


hooked_input = StringBuffer("hello world").with_input_hooks(MyInputHook())
hooked_output = StringBuffer().with_output_hooks(MyOutputHook())


def hello_world(input_data: str) -> str:
    print(input_data.upper())
    return input_data[::-1]


run(node(hello_world, inputs=hooked_input, outputs=hooked_output))
print(hooked_output.load())



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ordeq.framework.nodes import Node

from collections.abc import Hashable
from typing import Generic, TypeVar

T = TypeVar("T")


@dataclass(frozen=True)
class Registry(Generic[T]):
    """Class representing a registry of objects."""

    _data: dict[Hashable, T] = field(default_factory=dict, init=False)

    def get(self, key: Hashable) -> T:
        """Retrieve the value associated with the given key.

        Args:
            key: The key to look up in the registry.

        Returns:
            The value associated with the key.

        Raises:
            KeyError: If the key is not found in the registry.
        """  # noqa: DOC502 (Raised exception is not explicitly raised)

        return self._data[key]

    def set(self, key: Hashable, value: T) -> None:
        """Add a new key-value pair to the registry.

        Args:
            key: The key to add to the registry.
            value: The value to associate with the key.

        Raises:
            KeyError: If the key already exists in the registry.
        """

        if key in self._data:
            raise KeyError(f"Key '{key}' already exists")
        self._data[key] = value

    def __contains__(self, key: Hashable) -> bool:
        """Check if the item exists in the registry.

        Args:
            key: The key to check for existence.

        Returns:
            True if the key exists, False otherwise.
        """

        return key in self._data


# Global node registry used by Ordeq. Users should not use this directly.
NODE_REGISTRY: Registry[Node] = Registry()



from dataclasses import dataclass
from pathlib import Path

from ordeq import Output


@dataclass(kw_only=True, frozen=True)
class ExampleOutputSaveKwarg(Output):
    path: Path
    attribute: str

    def save(self, data: str, hello: str = "Guten tag") -> None:
        assert data == f"{self.path}@{self.attribute}: {hello} world!"


example_output = ExampleOutputSaveKwarg(path=Path("hello.txt"), attribute="L1")
# No kwarg:
example_output.save("hello.txt@L1: Guten tag world!")
# Alternative kwarg:
example_output.save("hello.txt@L1: Bonjour world!", hello="Bonjour")
print(type(example_output))

with_options = example_output.with_save_options(hello="Hello")
# No kwarg, with save options:
with_options.save("hello.txt@L1: Hello world!")
# Alternative kwarg, with save options:
with_options.save("hello.txt@L1: Buenos dias world!", hello="Buenos dias")
print(type(with_options))


@dataclass(kw_only=True, frozen=True)
class ExampleOutputSaveArg(Output):
    path: Path
    attribute: str

    def save(self, data: str, hello: str = "Hello") -> None:
        assert data == f"{self.path}@{self.attribute}: {hello} world!"


example_input_arg = ExampleOutputSaveArg(path=Path("hello.txt"), attribute="L1")
example_input_arg.save("hello.txt@L1: Hello world!", "Hello")
print(type(example_input_arg))

with_options_arg = example_input_arg.with_save_options(hello="Hello")
# This should raise a type error,
# but still run as we fill the missing arg on save:
with_options_arg.save("hello.txt@L1: Hello world!")
# Alternative arg, with save options:
with_options_arg.save("hello.txt@L1: Hi world!", hello="Hi")
print(type(with_options_arg))

example_input_arg.with_save_options(unknown_kwarg="Hello")  # should error



from ordeq import node, IO
from ordeq.framework.runner import run
from ordeq_common import Static, StringBuffer


class CustomIO(IO[str]):
    def __init__(self, attr: str = ""):
        self.attr = attr
        super().__init__()

    def load(self, suffix: str = "") -> str:
        return f"{self.attr} {suffix}"

    def save(self, value: str, suffix: str = "") -> None:
        self.attr += f"{value} {suffix}"

    def __repr__(self):
        return f"CustomIO(attr={self.attr})"


x1 = CustomIO("y did it")
x2 = CustomIO().with_save_options(
    suffix="!"
)
x3 = CustomIO("x did it").with_load_options(
    suffix="and I know the murder weapon"
)
x4 = CustomIO()


@node(inputs=x1, outputs=x2)
def increment(x: str) -> str:
    return f"x says {x}"


@node(inputs=[x2, x3], outputs=x4)
def decrement(x: str, y: str) -> str:
    return f"x says '{x}' but y says '{y}'"


result = run(increment, decrement, verbose=True)
print(result)



from dataclasses import dataclass

from ordeq import Output


@dataclass(kw_only=True, frozen=True)
class ExampleOutputsaveArg(Output):
    def save(self, data: str, hello: str = "...") -> None:
        pass


_ = ExampleOutputsaveArg().with_save_options(world="hello world")



from ordeq import Input
from resources.catalog.catalogs.remote import *  # 'remote' is the catalog that is overridden

from ordeq_common import Static

hello: Input[str] = Static("Hey I am overriding the hello IO")  # this overrides the base catalog



import logging
from collections.abc import Callable, Hashable, Sequence
from itertools import chain
from types import ModuleType
from typing import Literal, TypeVar, overload

from ordeq.framework.graph import NodeGraph
from ordeq.framework.hook import NodeHook, RunHook
from ordeq.framework.io import Input, Output, _InputCache
from ordeq.framework.nodes import NODE_REGISTRY, Node, get_node

logger = logging.getLogger(__name__)

T = TypeVar("T")

DataStoreType = dict[Input[T] | Output[T], T]
# The save mode determines which outputs are saved. When set to:
# - 'all', all outputs are saved, including those of intermediate nodes.
# - 'sinks', only outputs of sink nodes are saved, i.e. those w/o successors.
# - 'none', to dry-run and save no outputs
# Future extension:
# - 'last', which saves the output of the last node for which no error
# occurred. This can be useful for debugging.
SaveMode = Literal["all", "sinks", "none"]


def _save_outputs(
    node: Node, values: Sequence[T], save: bool = True
) -> dict[Input[T] | Output[T], T]:
    computed: dict[Input[T] | Output[T], T] = {}
    for output_dataset, data in zip(node.outputs, values, strict=False):
        computed[output_dataset] = data

        # TODO: this can be handled in the `save_wrapper`
        if save:
            output_dataset.save(data)

    return computed


def _run_node(
    node: Node, *, hooks: Sequence[NodeHook] = (), save: bool = True
) -> DataStoreType:
    node.validate()

    for node_hook in hooks:
        node_hook.before_node_run(node)

    args = [input_dataset.load() for input_dataset in node.inputs]

    # persisting loaded data
    for node_input, data in zip(node.inputs, args, strict=True):
        if isinstance(node_input, _InputCache):
            node_input.persist(data)

    logger.info("Running node %s", node)

    try:
        values = node.func(*args)
    except Exception as exc:
        for node_hook in hooks:
            node_hook.on_node_call_error(node, exc)
        raise exc

    if len(node.outputs) == 0:
        values = ()
    elif len(node.outputs) == 1:
        values = (values,)
    else:
        values = tuple(values)

    computed = _save_outputs(node, values, save=save)

    # persisting computed data only if outputs are loaded again later
    for node_output in node.outputs:
        if isinstance(node_output, _InputCache):
            node_output.persist(computed[node_output])  # ty: ignore[call-non-callable]

    for node_hook in hooks:
        node_hook.after_node_run(node)

    return computed


def _run_graph(
    graph: NodeGraph,
    *,
    hooks: Sequence[NodeHook] = (),
    save: SaveMode = "all",
    io: dict[Input[T] | Output[T], Input[T] | Output[T]] | None = None,
) -> DataStoreType:
    """Runs nodes in a graph topologically, ensuring IOs are loaded only once.

    Args:
        graph: node graph to run
        hooks: hooks to apply
        hooks: hooks to apply
        save: 'all' | 'sinks' | 'none'.
            If 'sinks', only saves the outputs of sink nodes in the graph.
        io: mapping of IO objects to their replacements

    Returns:
        a dict mapping each IO to the computed data

    """

    patched_nodes: dict[Node, Node] = {}
    for node in graph.nodes:
        patched_nodes[node] = _patch_io(node, io or {})

    data_store: dict = {}  # For each IO, the loaded data

    # TODO: Create _Patch wrapper for IO?
    for node in graph.topological_ordering:
        if (save == "sinks" and node in graph.sink_nodes) or save == "all":
            save_node = True
        else:
            save_node = False

        computed = _run_node(patched_nodes[node], hooks=hooks, save=save_node)
        data_store.update(computed)

    reverse_io = {v: k for k, v in (io or {}).items()}
    patched_data_store = {}
    for k, v in data_store.items():
        patched_data_store[reverse_io.get(k, k)] = v

    # unpersist IO objects
    for gnode in graph.nodes:
        io_objs = chain(gnode.inputs, gnode.outputs)
        for io_obj in io_objs:
            if isinstance(io_obj, _InputCache):
                io_obj.unpersist()

    return patched_data_store


def _patch_io(
    node: Node, io: dict[Input[T] | Output[T], Input[T] | Output[T]]
) -> Node:
    """Patches the inputs and outputs of a node with the provided IO mapping.

    Args:
        node: the original node
        io: mapping of Input/Output objects to their replacements

    Returns:
        the node with patched inputs and outputs

    """

    return node._replace(
        inputs=[io.get(ip, ip) for ip in node.inputs],  # type: ignore[misc]
        outputs=[io.get(op, op) for op in node.outputs],  # type: ignore[misc]
    )


def _gather_nodes_from_module(module: ModuleType) -> list[Node]:
    """Gathers all nodes defined in a module.

    Args:
        module: the module to gather nodes from

    Returns:
        a list of nodes defined in the module

    """

    nodes = []
    for attr in dir(module):
        obj = getattr(module, attr)
        if isinstance(obj, Hashable) and obj in NODE_REGISTRY:
            nodes.append(NODE_REGISTRY.get(obj))
    return nodes


@overload
def run(
    *runnables: ModuleType,
    hooks: Sequence[NodeHook | RunHook] = (),
    save: SaveMode = "all",
    verbose: bool = False,
    io: dict[Input[T] | Output[T], Input[T] | Output[T]] | None = None,
) -> DataStoreType: ...


@overload
def run(
    *runnables: Callable,
    hooks: Sequence[NodeHook | RunHook] = (),
    save: SaveMode = "all",
    verbose: bool = False,
    io: dict[Input[T] | Output[T], Input[T] | Output[T]] | None = None,
) -> DataStoreType: ...


def run(
    *runnables: ModuleType | Callable,
    hooks: Sequence[NodeHook | RunHook] = (),
    save: SaveMode = "all",
    verbose: bool = False,
    io: dict[Input[T] | Output[T], Input[T] | Output[T]] | None = None,
) -> DataStoreType:
    """Runs nodes in topological order.

    Args:
        runnables: the nodes to run, or a modules containing nodes
        hooks: hooks to apply
        save: 'all' | 'sinks'. If 'sinks', only saves the sink outputs
        verbose: whether to print the node graph
        io: mapping of IO objects to their replacements

    Returns:
        a dict mapping each IO to the computed data

    Raises:
        TypeError: if runnables are not all modules or all callables
    """

    if all(isinstance(r, ModuleType) for r in runnables):
        modules: tuple[ModuleType, ...] = runnables  # type: ignore[assignment]
        nodes_ = []
        for module in modules:
            nodes_.extend(_gather_nodes_from_module(module))
    elif all(callable(r) for r in runnables):
        funcs: tuple[Callable, ...] = runnables  # type: ignore[assignment]
        nodes_ = [get_node(func) for func in funcs]
    else:
        raise TypeError("All runnables must be either modules or nodes.")

    graph = NodeGraph.from_nodes(nodes_)

    if verbose:
        print(graph)

    node_hooks: list[NodeHook] = [h for h in hooks if isinstance(h, NodeHook)]
    run_hooks: list[RunHook] = [h for h in hooks if isinstance(h, RunHook)]

    for run_hook in run_hooks:
        run_hook.before_run(graph)

    result = _run_graph(graph, hooks=node_hooks, save=save, io=io)

    for run_hook in run_hooks:
        run_hook.after_run(graph, result)

    return result



from dataclasses import dataclass

import pytest
from ordeq import IO, Input, IOException, Output


def test_standalone_input():
    i, j = Input(), Input()
    with pytest.raises(IOException):
        i.load()

    assert repr(i) == f"Input(idx={i._idx})"
    assert i != j
    assert i == i  # noqa: PLR0124
    assert j == j  # noqa: PLR0124
    assert i._idx != j._idx


def test_standalone_output():
    k, x = Output(), Output()
    k.save("test")  # does nothing

    assert repr(k) == f"Output(idx={k._idx})"
    assert k != x
    assert k == k  # noqa: PLR0124
    assert x == x  # noqa: PLR0124
    assert k._idx != x._idx


def test_standalone_io():
    y, z = IO(), IO()
    with pytest.raises(IOException):
        y.load()
    y.save("test")  # does nothing

    assert repr(y) == f"IO(idx={y._idx})"
    assert y != z
    assert y == y  # noqa: PLR0124
    assert z == z  # noqa: PLR0124
    assert y._idx != z._idx


@dataclass(frozen=True)
class CustomDataclassIO(IO[str]):
    attr: str

    def load(self, extra: str = "") -> str:
        return self.attr + extra

    def save(self, value: str, suffix: str = ""):
        self.attr += value + suffix


class CustomIO(IO[str]):
    def __init__(self, attr: str):
        self.attr = attr
        super().__init__()

    def load(self, extra: str = "") -> str:
        return self.attr + extra

    def save(self, value: str, suffix: str = ""):
        self.attr += value + suffix


@pytest.mark.parametrize("io", [CustomIO("attr"), CustomDataclassIO("attr")])
def test_load_save_options(io: IO):
    # All IO instances need to be hashable
    # TODO: Check hashability in a metaclass
    # TODO: Also check for existence of a __repr__

    hash(io)
    io_wlo = io.with_load_options(extra="extra")
    hash(io_wlo)
    io_wso = io.with_save_options(suffix="_suffix")
    hash(io_wso)



from ordeq import node, run

from resources.catalog.catalogs import remote_extended

catalog = remote_extended


@node(inputs=catalog.hello, outputs=catalog.another_io)
def func1(hello: str) -> str:
    return f"{hello.upper()}!"


print(run(func1))



from ordeq import IO, Input, Output, node
from ordeq.framework.runner import run
from ordeq_common import StringBuffer

I1 = Input[str]()
I2 = Input[str]()
O1 = IO[str]()
O2 = Output[str]()


@node(inputs=[I1, I2], outputs=O1)
def f(i: str, j: str) -> str:
    return f"{i} {j}"


@node(inputs=O1, outputs=O2)
def g(a: str) -> str:
    return f(a, a)


print(run(f, g, verbose=True))  # raises NotImplementedError



from __future__ import annotations

import copy
import inspect
import logging
from collections.abc import Callable, Sequence
from functools import cached_property, reduce, wraps
from typing import Any, Generic, TypeVar
from uuid import uuid4

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

from ordeq.framework.hook import InputHook, OutputHook

logger = logging.getLogger(__name__)


class IOException(Exception):
    """Exception raised by IO implementations in case of load/save failure.
    IO implementations should provide instructive information.
    """


T = TypeVar("T")
Tin = TypeVar("Tin")
Tout = TypeVar("Tout")


def _find_references(attributes) -> dict[str, list[Input | Output | IO]]:
    """Find all attributes of type Input, Output, or IO.

    Args:
        attributes: a dictionary of attributes to inspect

    Returns:
        a dictionary mapping attribute names to lists of Input, Output, or IO
    """

    def get_io_instance(value: Any) -> list[Input | Output | IO]:
        if isinstance(value, (Input, Output, IO)):
            return [value]
        if isinstance(value, Sequence) and not isinstance(value, (str, bytes)):
            return [io for v in value for io in get_io_instance(v)]
        if isinstance(value, dict):
            return [io for v in value.values() for io in get_io_instance(v)]
        return []

    wrapped = {}
    for attribute, value in attributes.items():
        ios = get_io_instance(value)
        if ios:
            wrapped[attribute] = ios
    return wrapped


def _raise_not_implemented(*args, **kwargs):
    raise NotImplementedError()


def _load_decorator(load_func):
    @wraps(load_func)
    def wrapper(self, *args, **kwargs):
        # wrappers defined in the base classes
        # similar to super().load_wrapper() calls, without requiring
        # the `load_wrappers` to call each super.
        wrappers = [
            base.load_wrapper
            for base in reversed(type(self).__mro__)
            if hasattr(base, "load_wrapper")
        ]

        def base_func(*a, **k):
            logger.info("Loading %s", self)

            return load_func(self, *a, **k)

        composed = reduce(
            lambda prev_func, wrap: lambda *a, **k: wrap(
                self, prev_func, *a, **k
            ),
            wrappers,
            base_func,
        )
        return composed(*args, **kwargs)

    return wrapper


class _InputMeta(type):
    def __new__(cls, name, bases, class_dict):
        # Retrieve the closest load method
        load_method = _raise_not_implemented
        for base in bases:
            l_method = getattr(base, "load", None)
            if (
                l_method is not None
                and l_method.__qualname__ != "_raise_not_implemented"
            ):
                load_method = l_method

        l_method = class_dict.get("load", None)
        if (
            l_method is not None
            and l_method.__qualname__ != "_raise_not_implemented"
        ):
            load_method = l_method

        if name not in {"Input", "IO"}:
            # Ensure load method is implemented
            if (
                not callable(load_method)
                or load_method.__qualname__ == "_raise_not_implemented"
            ):
                msg = (
                    f"Can't instantiate abstract class {name} "
                    "with abstract method load"
                )
                raise TypeError(msg)

            # Ensure all arguments (except self/cls) have default values
            sig = inspect.signature(load_method)
            for argument, param in sig.parameters.items():
                if argument in {"self", "cls"}:
                    continue
                if (
                    param.default is inspect.Parameter.empty
                    and param.kind != inspect._ParameterKind.VAR_KEYWORD  # noqa: SLF001
                ):
                    raise TypeError(
                        f"Argument '{argument}' of function "
                        f"'{load_method.__name__}' has no default value."
                    )

        if not hasattr(load_method, "__wrapped__"):
            class_dict["load"] = _load_decorator(load_method)
        return super().__new__(cls, name, bases, class_dict)


class _BaseInput(Generic[Tin]):
    load: Callable = _raise_not_implemented


class _InputOptions(_BaseInput[Tin]):
    """Class that adds load options to an Input.
    Used for compartmentalizing load options, no reuse."""

    _load_options: dict[str, Any] | None = None

    def with_load_options(self, **load_options) -> Self:
        """Creates a new instance of self with load options set to kwargs.

        Note:
            the instance is shallow-copied. The new instance still references
            the attributes of the original instance.

        Returns:
            a new instance, with load options set to kwargs
        """

        new_instance = copy.copy(self)

        # ensure the `load_options` are valid for the `load` method
        inspect.signature(new_instance.load).bind_partial(**load_options)

        # Set the dict directly to support IO that are frozen dataclasses:
        new_instance.__dict__["_load_options"] = load_options
        return new_instance

    def load_wrapper(self, load_func, *args, **kwargs) -> Tin:
        # Compose load options and pass to the load_func
        load_options = self._load_options or {}

        # Kwargs take priority of load_options
        load_options.update(kwargs)
        return load_func(*args, **load_options)


class _InputHooks(_BaseInput[Tin]):
    """Class that adds input hooks to an Input.
    Used for compartmentalizing load options, no reuse."""

    input_hooks: tuple[InputHook, ...] = ()

    def with_input_hooks(self, *hooks: InputHook) -> Self:
        for hook in hooks:
            if not (
                isinstance(hook, InputHook) and not isinstance(hook, type)
            ):
                raise TypeError(f"Expected InputHook instance, got {hook}.")

        new_instance = copy.copy(self)
        new_instance.__dict__["input_hooks"] = hooks
        return new_instance

    def load_wrapper(self, load_func, *args, **kwargs) -> Tin:
        for hook in self.input_hooks:
            hook.before_input_load(self)  # type: ignore[arg-type]

        result = load_func(*args, **kwargs)

        for hook in self.input_hooks:
            hook.after_input_load(self, result)  # type: ignore[arg-type]

        return result


class _InputReferences(_BaseInput[Tin]):
    """Class that adds reference tracking to an Input.
    Used for compartmentalizing reference tracking, no reuse."""

    @cached_property
    def references(self) -> dict[str, list[Input | Output | IO]]:
        """Find all attributes of type Input, Output, or IO on the object.

        Returns:
            a dictionary mapping attribute names to lists of Input, Output,
            or IO
        """
        return _find_references(self.__dict__)


class _InputCache(_BaseInput[Tin]):
    """Class that adds caching to an Input."""

    _data: Tin

    def load_wrapper(self, load_func, *args, **kwargs) -> Tin:
        if not hasattr(self, "_data"):
            return load_func(*args, **kwargs)
        return self._data

    def persist(self, data: Tin) -> None:
        self.__dict__["_data"] = data

    def unpersist(self) -> None:
        if "_data" in self.__dict__:
            del self.__dict__["_data"]


class _InputException(_BaseInput[Tin]):
    def load_wrapper(self, load_func, *args, **kwargs) -> Tin:
        try:
            return load_func(*args, **kwargs)
        except Exception as exc:
            msg = f"Failed to load {self!s}.\n{exc!s}"
            raise IOException(msg) from exc


class Input(
    _InputOptions[Tin],
    _InputHooks[Tin],
    _InputReferences[Tin],
    _InputCache[Tin],
    _InputException[Tin],
    Generic[Tin],
    metaclass=_InputMeta,
):
    """Base class for all inputs in Ordeq. An `Input` is a class that loads
    data. All `Input` classes should implement a load method. By default,
    loading an input raises a `NotImplementedError`. See the Ordeq IO packages
    for some out-of-the-box implementations (e.g., `Static`, `StringBuffer`,
    etc.).

    `Input` can also be used directly as placeholder. This can be useful when
    you are defining a node, but you do not want to provide an actual input
    yet. In this case, you can:

    ```python
    >>> from ordeq.framework import Input, node
    >>> from ordeq_common import StringBuffer

    >>> name = Input[str]()
    >>> greeting = StringBuffer()

    >>> @node(
    ...     inputs=name,
    ...     outputs=greeting
    ... )
    ... def greet(name: str) -> str:
    ...     return f"Hello, {name}!"

    ```

    In the example above, `name` represents the placeholder input to the node
    `greet`. Running the node greet as-is will raise a `NotImplementedError`:

    ```python
    >>> from ordeq import run
    >>> run(greet) # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    NotImplementedError:

    ```

    To use the `greet` node, we need to provide an actual input. For instance:

    ```python
    >>> from ordeq_common import Static
    >>> result = run(greet, io={name: Static("Alice")})
    >>> result[greeting]
    'Hello, Alice!'
    ```
    """

    def __init__(self):
        self._idx = str(uuid4())

    def __repr__(self):
        return f"Input(idx={self._idx})"


def _save_decorator(save_func):
    @wraps(save_func)
    def wrapper(self, data, /, *args, **kwargs):
        # wrappers defined in the base classes
        # similar to super().save_wrapper() calls, without requiring
        # the `save_wrapper` to call each super.
        wrappers = [
            base.save_wrapper
            for base in reversed(type(self).__mro__)
            if hasattr(base, "save_wrapper")
        ]

        def base_func(d, *a, **k):
            logger.info("Saving %s", self)

            save_func(self, d, *a, **k)

        composed = reduce(
            lambda prev_func, wrap: lambda d, *a, **k: wrap(
                self, prev_func, d, *a, **k
            ),
            wrappers,
            base_func,
        )
        composed(data, *args, **kwargs)

    return wrapper


class _OutputMeta(type):
    def __new__(cls, name, bases, class_dict):
        # Retrieve the closest save method
        save_method = _raise_not_implemented
        for base in bases:
            s_method = getattr(base, "save", None)
            if s_method is not None and s_method.__qualname__ != "_pass":
                save_method = s_method

        s_method = class_dict.get("save", None)
        if s_method is not None and s_method.__qualname__ != "_pass":
            save_method = s_method

        if name not in {"Output", "IO"}:
            if not callable(save_method) or save_method == _pass:
                msg = (
                    f"Can't instantiate abstract class {name} "
                    "with abstract method save"
                )
                raise TypeError(msg)

            sig = inspect.signature(save_method)
            if len(sig.parameters) < 2:
                raise TypeError("Save method requires a data parameter.")

            # Ensure all arguments (except the first two, self/cls and data)
            # have default values
            for i, (argument, param) in enumerate(sig.parameters.items()):
                # Skip self/cls and data
                if i < 2:
                    continue
                if (
                    param.default is inspect.Parameter.empty
                    and param.kind != inspect._ParameterKind.VAR_KEYWORD  # noqa: SLF001
                ):
                    raise TypeError(
                        f"Argument '{argument}' of function "
                        f"'{save_method.__name__}' has no default value."
                    )

            if (
                sig.return_annotation != inspect.Signature.empty
                and sig.return_annotation is not None
            ):
                raise TypeError("Save method must have return type None.")

            if not hasattr(save_method, "__wrapped__"):
                class_dict["save"] = _save_decorator(save_method)
        return super().__new__(cls, name, bases, class_dict)


def _pass(*args, **kwargs):
    return


class _BaseOutput(Generic[Tout]):
    save: Callable = _pass


class _OutputOptions(_BaseOutput[Tout], Generic[Tout]):
    """Class that adds save options to an Output.
    Used for compartmentalizing save options, no reuse."""

    _save_options: dict[str, Any] | None = None

    def with_save_options(self, **save_options) -> Self:
        """Creates a new instance of self with save options set to kwargs.

        Note:
            the instance is shallow-copied. The new instance still references
            the attributes of the original instance.

        Returns:
            a new instance, with save options set to kwargs
        """

        new_instance = copy.copy(self)

        # ensure the `save_options` are valid for the `save` method
        inspect.signature(new_instance.save).bind_partial(**save_options)

        # Set the dict directly to support IO that are frozen dataclasses
        new_instance.__dict__["_save_options"] = save_options
        return new_instance

    def save_wrapper(self, save_func, data: Tout, *args, **kwargs) -> None:
        save_options = self._save_options or {}

        # Kwargs take priority of save_options
        save_options.update(kwargs)
        save_func(data, *args, **save_options)


class _OutputHooks(_BaseOutput[Tout], Generic[Tout]):
    """Class that adds output hooks to an Output.
    Used for compartmentalizing load options, no reuse."""

    output_hooks: tuple[OutputHook, ...] = ()

    def with_output_hooks(self, *hooks: OutputHook) -> Self:
        for hook in hooks:
            if not (
                isinstance(hook, OutputHook) and not isinstance(hook, type)
            ):
                raise TypeError(f"Expected OutputHook instance, got {hook}.")

        new_instance = copy.copy(self)
        new_instance.__dict__["output_hooks"] = hooks
        return new_instance

    def save_wrapper(self, save_func, data: Tout, *args, **kwargs) -> None:
        for hook in self.output_hooks:
            hook.before_output_save(self, data)  # type: ignore[arg-type]

        save_func(data, *args, **kwargs)

        for hook in self.output_hooks:
            hook.after_output_save(self, data)  # type: ignore[arg-type]


class _OutputReferences(_BaseOutput[Tout], Generic[Tout]):
    """Class that adds reference tracking to an Output.
    Used for compartmentalizing reference tracking, no reuse."""

    @cached_property
    def references(self) -> dict[str, list[Input | Output | IO]]:
        """Find all attributes of type Input, Output, or IO on the object.

        Returns:
            a dictionary mapping attribute names to lists of Input, Output,
            or IO
        """
        return _find_references(self.__dict__)


class _OutputException(_BaseOutput[Tout]):
    def save_wrapper(self, save_func, data: Tout, *args, **kwargs) -> None:
        try:
            save_func(data, *args, **kwargs)
        except Exception as exc:
            msg = f"Failed to save {self!s}.\n{exc!s}"
            raise IOException(msg) from exc


class Output(
    _OutputOptions[Tout],
    _OutputHooks[Tout],
    _OutputReferences[Tout],
    _OutputException[Tout],
    Generic[Tout],
    metaclass=_OutputMeta,
):
    """Base class for all outputs in Ordeq. An `Output` is a class that saves
    data. All `Output` classes should implement a save method. By default,
    saving an output does nothing. See the Ordeq IO packages for some
    out-of-the-box implementations (e.g., `YAML`, `StringBuffer`, etc.).

    `Output` can also be used directly as placeholder. This can be useful when
    you are defining a node, but you do not want to provide an actual output.
    In this case, you can:

    ```python
    >>> from ordeq.framework import Output, node
    >>> from ordeq_common import StringBuffer

    >>> greeting = StringBuffer("hello")
    >>> greeting_upper = Output[str]()

    >>> @node(
    ...     inputs=greeting,
    ...     outputs=greeting_upper
    ... )
    ... def uppercase(greeting: str) -> str:
    ...     return greeting.upper()

    ```

    In the example above, `greeting_upper` represents the placeholder output
    to the node `uppercase`. When you run the node `uppercase`, its result can
    be retrieved from the `greeting_upper` output. For instance:

    ```python
    >>> from ordeq import run
    >>> result = run(uppercase)
    >>> result[greeting_upper]
    'HELLO'
    ```
    """

    def __init__(self):
        self._idx = str(uuid4())

    def __repr__(self):
        return f"Output(idx={self._idx})"


class _IOMeta(_InputMeta, _OutputMeta): ...


class IO(Input[T], Output[T], metaclass=_IOMeta):
    """Base class for all IOs in Ordeq. An `IO` is a class that can both load
    and save data. See the Ordeq IO packages for some out-of-the-box
    implementations (e.g., `YAML`, `StringBuffer`, etc.).

    `IO` can also be used directly as placeholder. This can be useful when
    you want to pass data from one node to another, but you do not want to save
    the data in between:

    ```python
    >>> from ordeq import Input, node
    >>> from ordeq_common import StringBuffer, Static

    >>> hello = StringBuffer("hi")
    >>> name = Static("Bob")
    >>> greeting = IO[str]()
    >>> greeting_capitalized = StringBuffer()

    >>> @node(
    ...     inputs=[hello, name],
    ...     outputs=greeting
    ... )
    ... def greet(greeting: str, name: str) -> str:
    ...     return f"{greeting}, {name}!"

    >>> @node(
    ...     inputs=greeting,
    ...     outputs=greeting_capitalized
    ... )
    ... def capitalize(s: str) -> str:
    ...     return s.capitalize()
    ```

    In the example above, `greeting` represents the placeholder output
    to the node `greet`, as well as the placeholder input to `capitalize`.

    When you run the nodes `greeting` and `capitalize` the result of `greeting`
    will be passed along unaffected to `capitalize`, much like a cache:

    ```python
    >>> from ordeq import run
    >>> result = run(greet, capitalize)
    >>> result[greeting]
    'hi, Bob!'
    ```
    """

    def __repr__(self):
        return f"IO(idx={self._idx})"



from dataclasses import dataclass

from ordeq import Output


@dataclass(kw_only=True, frozen=True)
class ExampleOutputNosave(Output):
    def save(self) -> None:
        pass


_ = ExampleOutputNosave()



import pytest
from ordeq import Node, node
from ordeq.framework.pipeline import Pipeline, is_pipeline
from ordeq_common import StringBuffer

node_1 = node(inputs=(), outputs=(), func=lambda: None)
node_2 = node(inputs=(), outputs=(StringBuffer(),), func=lambda: "bluh")


@pytest.mark.parametrize(
    ("obj", "expected"),
    [
        (object, False),
        ({}, False),
        ({1, 2, 3}, False),
        ({"a", 1}, False),
        ({"a", node_1}, False),
        ({node_1}, True),
        ({node_1, node_2}, True),
        (Pipeline, False),
        (Node, False),
    ],
)
def test_is_pipeline(obj: object, expected: bool):
    assert is_pipeline(obj) == expected



from ordeq import Node, NodeHook, Output, OutputHook, node
from ordeq.framework import get_node
from ordeq_common import StringBuffer


class MyUntypedOutputHook(OutputHook):
    def before_output_save(
        self, op: Output[str], data: str
    ) -> None:
        print(f"saving data `{data}` to output `{op}`")


class MyFixedOutputHook(MyUntypedOutputHook, NodeHook):
    def before_node_run(self, node: Node) -> None:
        print(f"running node {node}")


@node(inputs=StringBuffer("a"), outputs=StringBuffer("b"))
def func(x: str) -> str:
    return x


# This is just to ensure the hook can be instantiated without errors.
untyped_hook = MyUntypedOutputHook()
untyped_hook.before_output_save(StringBuffer("A"), "hello")

fixed_output_hook = MyFixedOutputHook()
fixed_output_hook.before_node_run(get_node(func))
fixed_output_hook.before_output_save(StringBuffer("B"), "world")



from collections.abc import Iterable
from graphlib import TopologicalSorter

from ordeq.framework.nodes import Node

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self
EdgesType = dict[Node, list[Node]]


def _build_graph(nodes: Iterable[Node]) -> EdgesType:
    """Builds a mapping of node to node(s), i.e., the edge map of a graph.

    Args:
        nodes: iterable of `Node` objects

    Returns:
        a mapping of node to node(s), i.e., the edge map of a graph

    Raises:
        ValueError: if an output is defined by more than one node
    """
    output_to_node: dict = {}
    input_to_nodes: dict = {}
    edges: dict = {node: [] for node in nodes}
    for node in nodes:
        for output_ in node.outputs:
            if output_ in output_to_node:
                msg = f"IO {output_} cannot be outputted by more than one node"
                raise ValueError(msg)
            output_to_node[output_] = node
        for input_ in node.inputs:
            input_to_nodes[input_] = [*input_to_nodes.get(input_, []), node]
    for node_output, node in output_to_node.items():
        if node_output in input_to_nodes:
            edges[node] += input_to_nodes[node_output]
    return edges


def _find_topological_ordering(edges: EdgesType) -> tuple[Node, ...]:
    """Topological sort.

    Args:
        edges: mapping of node to node(s), i.e., the edge map of a graph

    Returns:
            a tuple of nodes in topological order
    """
    return tuple(reversed(tuple(TopologicalSorter(edges).static_order())))


def _find_sink_nodes(edges: EdgesType) -> set[Node]:
    """Finds the sinks nodes, i.e. nodes without successors.

    Args:
        edges: the graph

    Returns:
        set of the sink nodes

    """

    return {s for s, targets in edges.items() if len(targets) == 0}


def _nodes(edges: EdgesType) -> set[Node]:
    """Returns the set of all nodes.

    Args:
        edges: the graph

    Returns:
        set of all nodes
    """

    return set(edges.keys())


class NodeGraph:
    def __init__(self, edges: EdgesType):
        self.edges = edges

    @classmethod
    def from_nodes(cls, nodes: Iterable[Node]) -> Self:
        return cls(_build_graph(nodes))

    @property
    def topological_ordering(self) -> tuple[Node, ...]:
        return _find_topological_ordering(self.sorted_edges)

    @property
    def sorted_edges(self) -> EdgesType:
        return dict(
            sorted(
                [
                    (node, sorted(targets, key=lambda n: n.name))
                    for node, targets in self.edges.items()
                ],
                key=lambda x: x[0].name,
            )
        )

    @property
    def sink_nodes(self) -> set[Node]:
        """Finds the sink nodes, i.e., nodes without successors.

        Returns:
            set of the sink nodes
        """
        return _find_sink_nodes(self.edges)

    @property
    def nodes(self) -> set[Node]:
        """Returns the set of all nodes in this graph.

        Returns:
            all nodes in this graph
        """

        return _nodes(self.edges)

    def __repr__(self) -> str:
        lines = ["NodeGraph:", "  Edges:"]
        for node, targets in self.sorted_edges.items():
            targets_str = ", ".join(t.name for t in targets)
            lines.append(f"     {node.name} -> [{targets_str}]")
        lines.append("  Nodes:")
        lines.extend(f"     {node!r}" for node in self.sorted_edges)
        return "\n".join(lines)



from ordeq import Node, Output, OutputHook


class MyTypedOutputHook(OutputHook[str]):
    def before_output_save(
        self, op: Output[str], data: str
    ) -> None:
        print(f"saving data `{data}` to output `{op}`")

_ = MyTypedOutputHook()



from dataclasses import dataclass, field
from typing import Generator, Iterable

from ordeq import node, IO
from ordeq.framework.runner import run
from ordeq_common import Static


@dataclass(eq=False)
class Stream(IO[Generator[str, None, None]]):
    data: Iterable[str] = field(default_factory=list, hash=False)

    def load(self) -> Generator[str, None, None]:
        for item in self.data:
            yield item

    def save(self, data: Generator[str, None, None]) -> None:
        for item in data:
            self.data += [item]


x1 = Stream(["1", "2", "3"])
x2 = Stream()
x3 = Static("2")
x4 = Stream()


@node(inputs=x1, outputs=x2)
def increment(items: Generator[str, None, None]) -> Generator[str, None, None]:
    for item in items:
        yield str(int(item) + 1)


@node(inputs=[x2, x3], outputs=x4)
def multiply(items: Generator[str, None, None], y: str) -> Generator[
    str, None, None
]:
    for item in items:
        yield str(int(item) * int(y))


# Saving regularly yields no data in x4, since the stream of x2 in x2.save:
run(increment, multiply, verbose=True)

# Save using save="sinks" yields data in x4, but not in x2 (now, x2 is
# (exhausted in the for loop of multiply):
run(increment, multiply, verbose=True, save="sinks")



from collections.abc import Callable

from ordeq import Node, node
from ordeq.framework import get_node
from ordeq_common import StringBuffer

mock_x = StringBuffer("X")
mock_y = StringBuffer("Y")
mock_z = StringBuffer("Z")


@node(inputs=[mock_x], outputs=[mock_z])
def func(x: str = "X") -> Callable:
    @node(inputs=[mock_y], outputs=[mock_z])
    def inner(y: str) -> str:
        return x + y

    return inner


inner_func = func()
assert inner_func("X") == "XX"
assert get_node(func) == Node(func=func, inputs=(mock_x,), outputs=(mock_z,))
assert get_node(inner_func) == Node(
    func=inner_func, inputs=(mock_y,), outputs=(mock_z,)
)



import difflib
import importlib.util
import logging
import re
from pathlib import Path

import pytest
from mypy import api as mypy_api

PACKAGE_DIR = Path(__file__).resolve().parent.parent
RESOURCE_DIR = PACKAGE_DIR / "resources"
SNAPSHOT_DIR = PACKAGE_DIR / "snapshots"

FILES = list(RESOURCE_DIR.rglob("*.py"))


def _replace_pattern_with_seq(text: str, pattern: str, prefix: str) -> str:
    """Replace unique matches of pattern with prefix1, prefix2, ..."""
    seen = {}
    for match in re.finditer(pattern, text):
        val = match.group(0)
        if val not in seen:
            seen[val] = f"{prefix}{len(seen) + 1}"

    def repl(m):
        return seen[m.group(0)]

    return re.sub(pattern, repl, text)


def replace_object_hashes(text: str) -> str:
    """Replace object hashes like 0x103308890 with HASH1, HASH2, etc."""
    return _replace_pattern_with_seq(text, r"0x[0-9a-fA-F]+", "HASH")


def replace_uuid4(text: str) -> str:
    """Replace UUID4 strings with ID1, ID2, etc."""
    return _replace_pattern_with_seq(
        text,
        r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}",
        "ID",
    )


@pytest.mark.parametrize(
    "file_path",
    FILES,
    ids=[str(file.relative_to(RESOURCE_DIR)) for file in FILES],
)
def test_resources(file_path: Path, capsys, caplog) -> None:
    caplog.set_level(logging.INFO)
    caplog.handler.setFormatter(
        logging.Formatter(fmt="%(levelname)s\t%(name)s\t%(message)s")
    )

    snapshot_path = SNAPSHOT_DIR / file_path.relative_to(RESOURCE_DIR)
    snapshot_path.parent.mkdir(parents=True, exist_ok=True)
    snapshot_path = snapshot_path.with_suffix(".snapshot")

    # Dynamically import the module
    spec = importlib.util.spec_from_file_location(file_path.stem, file_path)
    module = importlib.util.module_from_spec(spec)

    sections = {}
    try:
        spec.loader.exec_module(module)
    except Exception as e:
        sections["Exception"] = f"{type(e).__name__}: {e}"

    captured_out_err = capsys.readouterr()
    if captured_out_err.out:
        sections["Output"] = captured_out_err.out
    if captured_out_err.err:
        sections["Error"] = captured_out_err.err
    if caplog.text:
        sections["Logging"] = caplog.text

    type_out, _, exit_code = mypy_api.run([str(file_path)])
    if exit_code != 0:
        sections["Typing"] = type_out

    output = "\n\n".join(
        f"{key}:\n{value.rstrip()}" for key, value in sections.items()
    )

    # Normalize object hashes
    captured = replace_uuid4(replace_object_hashes(output))

    # Read snapshot
    expected = (
        snapshot_path.read_text() if snapshot_path.exists() else "<NONE>"
    )
    expected = replace_object_hashes(expected)

    # Compare and write diff if not equal
    if captured != expected:
        diff = "\n".join(
            difflib.unified_diff(
                expected.splitlines(),
                captured.splitlines(),
                fromfile="expected",
                tofile="actual",
            )
        )
        snapshot_path.with_suffix(".snapshot").write_text(captured)
        pytest.fail(f"Output does not match snapshot:\n{diff}")



from ordeq import Output, node, run


class Example(Output[str]):
    def save(self, data: str) -> None:
        print("saving!", data)


example = Example()


@node(outputs=[example])
def my_node() -> str:
    return "Hello, World!"

@node(inputs=[example])
def load_node(data: str) -> None:
    print("loading!", data)


run(my_node, load_node)



import pytest
from ordeq import Node
from ordeq.framework._registry import Registry


def f():
    pass


def test_it_sets():
    registry = Registry()

    n = Node(func=f, inputs=(), outputs=())
    registry.set(f, n)
    assert registry._data[f] == n


def test_it_gets():
    registry = Registry()

    n = Node(func=f, inputs=(), outputs=())
    registry._data[f] = n
    assert registry.get(f) == n


def test_it_errors_when_set_twice():
    registry = Registry()

    n = Node(func=f, inputs=(), outputs=())
    registry.set(f, n)
    with pytest.raises(KeyError, match=r"Key '.+' already exists"):
        registry.set(f, n)


def test_it_contains():
    registry = Registry()

    n = Node(func=f, inputs=(), outputs=())
    registry.set(f, n)
    assert f in registry

    def unregistered(): ...

    assert unregistered not in registry



from ordeq import node
from ordeq.framework.runner import run
from ordeq_common import StringBuffer

I1 = StringBuffer("Hello")
I2 = StringBuffer("world!")


def f1(i: str, j: str) -> str:
    return f"{i} {j}"


def c():
    a = I1.load()
    b = I2.load()
    return f1(a, b)


print(c())

print(node(c, outputs=StringBuffer("result"))())

output = StringBuffer("result")
print(run(node(c, outputs=output))[output])



from ordeq import Output


class Example(Output[str]):
    def save(self, df: str) -> None:
        print("saving!", data)


data = "..."

example = Example()
example.save(data)  # ok
example.save(data=data)  # should give an error



from ordeq import node
from ordeq.framework.runner import run
from ordeq_common import StringBuffer

I1 = StringBuffer("Hello")
I2 = StringBuffer("world!")
R1 = StringBuffer()
R2 = StringBuffer()
R3 = StringBuffer()
R4 = StringBuffer()


@node(inputs=[I1, I2], outputs=[R1])
def f1(i: str, j: str) -> str:
    return f"{i} + {j}"


@node(inputs=[I2, R1], outputs=[R2])
def f2(i: str, j: str) -> str:
    return f"{i} - {j}"


@node(inputs=[R1], outputs=[R3])
def f3(i: str) -> str:
    return f"{i} * 2"


@node(inputs=[R1, R2, R3], outputs=[R4])
def f4(i: str, j: str, k: str) -> str:
    return f"{i} / {j} + {k}"


pipeline = {f1, f2, f3, f4}

output = run(*pipeline, save="all", verbose=True)
print(output[R4])

output = run(*pipeline, save="sinks", verbose=True)
print(output[R4])

output = run(*pipeline, save="none", verbose=True)
print(output[R4])



import copy

import pytest
from ordeq import Node, node
from ordeq.framework import get_node
from ordeq.framework.graph import NodeGraph
from ordeq.framework.runner import _run_graph, _run_node, run
from ordeq_common import StringBuffer

A, B, D = [StringBuffer(c) for c in "ABD"]
C = StringBuffer()
E = StringBuffer()
F = StringBuffer()


def test_run_regular_node():
    Bp = copy.copy(B)
    Bp.unpersist()
    node = Node(inputs=(A,), outputs=(Bp,), func=lambda x: x + x)
    computed = _run_node(node, hooks=())
    assert computed[Bp] == "AA"


def test_run_node_returns_none():
    node = Node(inputs=(A,), outputs=(), func=lambda _: None)
    computed = _run_node(node, hooks=())
    assert computed == {}


def test_run_node_with_zero_inputs():
    Bp = copy.copy(B)
    Bp.unpersist()
    node = Node(inputs=(), outputs=(Bp,), func=lambda: "something")
    loaded = {Bp: "something"}
    computed = _run_node(node, hooks=())
    assert loaded == computed


def test_run_node_with_zero_inputs_and_outputs():
    node = Node(inputs=(), outputs=(), func=lambda: print("Hello!"))
    loaded = {}
    computed = _run_node(node, hooks=())
    assert loaded == computed


plus = node(func=lambda x, y: f"{x} + {y}", inputs=(A, B), outputs=(C,))
minus = node(func=lambda x, y: f"{x} - {y}", inputs=(C, D), outputs=(E,))
square = node(func=lambda x: f"({x})^2", inputs=(E,), outputs=(F,))


@pytest.mark.parametrize(
    ("nodes", "expected_data_store"),
    [
        (
            [get_node(plus), get_node(minus), get_node(square)],
            {
                C: "A + BAAsomething",
                E: "A + BAAsomething - D",
                F: "(A + BAAsomething - D)^2",
            },
        ),
        (
            [get_node(plus), get_node(minus)],
            {C: "A + BAAsomething", E: "A + BAAsomething - D"},
        ),
        ([get_node(plus)], {C: "A + BAAsomething"}),
    ],
)
def test_run_graph(nodes, expected_data_store):
    data_store = _run_graph(NodeGraph.from_nodes(nodes))
    assert data_store == expected_data_store


@pytest.mark.parametrize(
    ("nodes", "expected_data_store"),
    [
        (
            (plus, minus, square),
            {
                C: "A + BAAsomething",
                E: "A + BAAsomething - D",
                F: "(A + BAAsomething - D)^2",
            },
        ),
        ((plus, minus), {C: "A + BAAsomething", E: "A + BAAsomething - D"}),
        ((plus,), {C: "A + BAAsomething"}),
    ],
)
def test_run_parametrized(nodes, expected_data_store):
    data_store = run(*nodes)
    assert data_store == expected_data_store



from ordeq import node, run
from ordeq_common import StringBuffer


@node(outputs=StringBuffer())
def func1() -> str:
    return "Hello"


@node(outputs=StringBuffer())
def func2() -> str:
    return "world"


run(func1, func2, verbose=True)



from ordeq import node
from ordeq_common import StringBuffer

node_1 = node(inputs=(), outputs=(), func=lambda: None)
node_2 = node(inputs=(), outputs=(StringBuffer(),), func=lambda: "bluh")

pipeline = {node_1, node_2}

# TODO: run these!



from ordeq import node
from ordeq.framework.catalog import check_catalogs_are_consistent

from resources.catalog.catalogs import inconsistent, local

check_catalogs_are_consistent(local, inconsistent)
catalog = inconsistent


@node(inputs=catalog.hello, outputs=catalog.result)
def func(hello: str) -> str:
    return f"{hello.upper()}!"



from ordeq import Output, NodeHook, OutputHook, Node, node, run


class Example(Output[str]):
    def save(self, data: str) -> None:
        print("saving!", data)


class MixedHook(NodeHook, OutputHook):
    node: Node | None = None

    def before_node_run(self, node: Node):
        self.node = node

    def before_output_save(self, output: Output, data) -> None:
        if self.node is not None:
            print(f"Hook: before saving output of node {self.node.name} with data: {data}")


hook = MixedHook()
example = Example().with_output_hooks(hook)

@node(outputs=[example])
def my_node() -> str:
    return "Hello, World!"


run(my_node, hooks=[hook])
